# Copyright © 2025 Mathieu Pasquet
# This file is part of slixmpp
# See the file LICENSE for copying permission.
import json
import logging
from datetime import datetime
from urllib.parse import urlparse

from slixmpp import JID
from slixmpp.plugins import BasePlugin, xep_0082
from dataclasses import dataclass

log = logging.getLogger(__name__)


class AiohttpNotFound(Exception):
    pass


@dataclass
class ExternalStatus:
    planned: bool | None
    beginning: datetime
    expected_end: datetime | None
    message: dict[str, str] | None


class XEP_0455(BasePlugin):
    """
    XEP-0455: Service Outage Status
    """
    name = 'xep_0455'
    description = 'XEP-0455: Service Outage Status'
    dependencies = {'xep_0128', 'xep_0030', 'xep_0082'}
    namespace = 'urn:xmpp:sos:0'

    async def get_external_status_addresses(self, domain: JID | None = None,
                                            **iqkwargs) -> list[str]:
        """Return the list of external status addresses for this domain.

        :param domain: Domain to disco to find a service.
        """
        if domain is None:
            domain = JID(self.xmpp.boundjid.host)

        uris = []
        results = await self.xmpp['xep_0030'].get_info(jid=domain,
            **iqkwargs
        )

        disco = results.get_plugin('disco_info', check=True)
        if disco is None or 'forms' not in disco:
            return uris
        forms = disco['forms']
        if not forms:
            return uris
        field = 'external-status-addresses'
        for form in forms:
            values = form.get_values()
            if values.get('FORM_TYPE') == [self.namespace]:
                uris.extend(values.get(field, []))
        return uris

    @classmethod
    async def fetch_status(cls, addresses: str | list[str]) -> dict:
        """
        Get the external status from a list of addresses.
        Only works with http/https for now and stops on the first status
        fetched successfully.

        :param addresses: address or list of addresses to fetch the status from.
        """
        try:
            from aiohttp import ClientSession
        except ImportError:
            raise AiohttpNotFound("aiohttp was not found, unable to download statuses")

        if not isinstance(addresses, list):
            addresses = [addresses]

        async with ClientSession(headers={'User-Agent': 'slixmpp'}) as session:
            for address in addresses:
                scheme, *_ = urlparse(address)
                if scheme not in ('http', 'https'):
                    continue
                response = await session.get(address, timeout=60)
                if response.status >= 400:
                    log.debug(f'Server "{address}" answered with code {response.status}')
                    continue
                text = await response.text()
                status = cls._parse_status(text)
                if status is not None:
                    return status

    @staticmethod
    def _parse_status(raw: str) -> ExternalStatus | None:
        """
        Parse a status json payload. Return None if the required (beginning)
        field is not found or not parseable.

        Returns a dataclass with the fields.
        """
        try:
            payload = json.loads(raw)
        except json.JSONDecodeError:
            log.error('Unable to parse the external status: {text}')
            return None
        if not payload:
            return None
        beginning_raw = payload.get('beginning')
        try:
            beginning = xep_0082.parse(beginning_raw)
        except (ValueError, TypeError):
            log.error(f'Bad value for beginning: "{beginning_raw}"')
            return None

        expected_end_raw = payload.get('expected_end')
        expected_end = None
        try:
            expected_end = xep_0082.parse(expected_end_raw)
        except (ValueError, TypeError):
            log.error(f'Bad value for expected end: "{expected_end_raw}"')

        planned_raw = payload.get('planned')
        planned = planned_raw if isinstance(planned_raw, bool) else None

        message_raw = payload.get('message')
        message = None
        if isinstance(message_raw, dict):
            message = {}
            for key, value in message_raw.items():
                if isinstance(key, str) and isinstance(value, str):
                    message[key] = value
        return ExternalStatus(planned, beginning, expected_end, message)



# Slixmpp: The Slick XMPP Library
# Copyright © 2026 nicoco <nicoco@nicoco.fr>
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.

"""
Usage:

>>> register_plugin(for_doctests=True)  # automatically done when enabling the XEP_0463 plugin
>>> presence = Presence()
>>> presence["muc_join"]["mav"]["since"] = "opaque-ver"
>>> presence.pretty_print()
<presence xmlns="jabber:client">
  <x xmlns="http://jabber.org/protocol/muc">
    <mav xmlns="urn:xmpp:muc:affiliations:1" since="opaque-ver" />
  </x>
</presence>

>>> msg = Message()
>>> msg["muc"]["mav"]["since"] = "since-ver"
>>> msg["muc"]["mav"]["until"] = "until-ver"
>>> msg.pretty_print()
<message xmlns="jabber:client">
  <x xmlns="http://jabber.org/protocol/muc#user">
    <mav xmlns="urn:xmpp:muc:affiliations:1" since="since-ver" until="until-ver" />
  </x>
</message>

>>> presence = Presence()
>>> presence["muc"]["mav"]["until"] = "XXX"
>>> presence.pretty_print()
<presence xmlns="jabber:client">
  <x xmlns="http://jabber.org/protocol/muc#user">
    <mav xmlns="urn:xmpp:muc:affiliations:1" until="XXX" />
  </x>
</presence>
"""

from slixmpp import Message, Presence
from slixmpp.plugins.xep_0045.stanza import MUCBase, MUCJoin, MUCPresence, MUCMessage
from slixmpp.xmlstream import ElementBase, register_stanza_plugin

NS = "urn:xmpp:muc:affiliations:1"


class Mav(ElementBase):
    namespace = NS
    name = "mav"
    plugin_attrib = "mav"
    interfaces = {"until", "since"}


def register_plugin(for_doctests=False) -> None:
    if for_doctests:
        register_stanza_plugin(Presence, MUCJoin)
        register_stanza_plugin(Presence, MUCBase)
        register_stanza_plugin(Presence, MUCPresence)
        register_stanza_plugin(Message, MUCMessage)
    register_stanza_plugin(MUCJoin, Mav)
    register_stanza_plugin(MUCBase, Mav)
    register_stanza_plugin(MUCMessage, Mav)
    register_stanza_plugin(MUCPresence, Mav)




# Slixmpp: The Slick XMPP Library
# Copyright (C) 2013 Nathanael C. Fritz, Lance J.T. Stout
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
import logging

from slixmpp.plugins import BasePlugin
from slixmpp.exceptions import XMPPError
from slixmpp.xmlstream import register_stanza_plugin
from slixmpp.plugins.xep_0048 import stanza, Bookmarks


log = logging.getLogger(__name__)


class XEP_0048(BasePlugin):

    name = 'xep_0048'
    description = 'XEP-0048: Bookmarks'
    dependencies = {'xep_0045', 'xep_0049', 'xep_0060', 'xep_0163', 'xep_0223'}
    stanza = stanza
    default_config = {
        'auto_join': False,
        'storage_method': 'xep_0049'
    }

    def plugin_init(self):
        register_stanza_plugin(self.xmpp['xep_0060'].stanza.Item, Bookmarks)

        self.xmpp['xep_0049'].register(Bookmarks)
        self.xmpp['xep_0163'].register_pep('bookmarks', Bookmarks)

        self.xmpp.add_event_handler('session_start', self._autojoin)

    def plugin_end(self):
        self.xmpp.del_event_handler('session_start', self._autojoin)

    def _autojoin(self, __):
        if not self.auto_join:
            return

        try:
            result = self.get_bookmarks(method=self.storage_method)
        except XMPPError:
            return

        if self.storage_method == 'xep_0223':
            bookmarks = result['pubsub']['items']['item']['bookmarks']
        else:
            bookmarks = result['private']['bookmarks']

        for conf in bookmarks['conferences']:
            if conf['autojoin']:
                log.debug('Auto joining %s as %s', conf['jid'], conf['nick'])
                self.xmpp['xep_0045'].join_muc(conf['jid'], conf['nick'],
                        password=conf['password'])

    def set_bookmarks(self, bookmarks, method=None, **iqargs):
        if not method:
            method = self.storage_method
        return self.xmpp[method].store(bookmarks, **iqargs)

    def get_bookmarks(self, method=None, **iqargs):
        if not method:
            method = self.storage_method

        loc = 'storage:bookmarks' if method == 'xep_0223' else 'bookmarks'

        return self.xmpp[method].retrieve(loc, **iqargs)




# Slixmpp: The Slick XMPP Library
# Copyright (C) 2012 Nathanael C. Fritz, Lance J.T. Stout
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.


from slixmpp import Iq
from slixmpp.plugins import BasePlugin
from slixmpp.xmlstream import register_stanza_plugin
from slixmpp.plugins.xep_0279 import stanza, IPCheck


class XEP_0279(BasePlugin):

    name = 'xep_0279'
    description = 'XEP-0279: Server IP Check'
    dependencies = {'xep_0030'}
    stanza = stanza

    def plugin_init(self):
        register_stanza_plugin(Iq, IPCheck)

    def session_bind(self, jid):
        self.xmpp['xep_0030'].add_feature('urn:xmpp:sic:0')

    def plugin_end(self):
        self.xmpp['xep_0030'].del_feature(feature='urn:xmpp:sic:0')

    def check_ip(self, ifrom=None, timeout=None, callback=None):
        iq = self.xmpp.Iq()
        iq['type'] = 'get'
        iq['from'] = ifrom
        iq.enable('ip_check')
        return iq.send(timeout=timeout, callback=callback)




# Slixmpp: The Slick XMPP Library
# Copyright (C) 2017 Emmanuel Gil Peyrot <linkmauve@linkmauve.fr>
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.

from slixmpp.stanza import Message
from slixmpp.plugins import BasePlugin
from slixmpp.xmlstream import register_stanza_plugin, ET, tostring
from slixmpp.plugins.xep_0394 import stanza, Markup, Span, BlockCode, List, Li, BlockQuote
from slixmpp.plugins.xep_0071 import XHTML_IM


class Start:
    def __init__(self, elem):
        self.elem = elem

    def __repr__(self):
        return 'Start(%s)' % self.elem


class End:
    def __init__(self, elem):
        self.elem = elem

    def __repr__(self):
        return 'End(%s)' % self.elem


class XEP_0394(BasePlugin):

    name = 'xep_0394'
    description = 'XEP-0394: Message Markup'
    dependencies = {'xep_0030', 'xep_0071'}
    stanza = stanza

    def plugin_init(self):
        register_stanza_plugin(Message, Markup)

    def session_bind(self, jid):
        self.xmpp['xep_0030'].add_feature(feature=Markup.namespace)

    def plugin_end(self):
        self.xmpp['xep_0030'].del_feature(feature=Markup.namespace)

    @staticmethod
    def _split_first_level(body, markup_elem):
        split_points = []
        elements = {}
        for markup in markup_elem['substanzas']:
            start = markup['start']
            end = markup['end']
            split_points.append(start)
            split_points.append(end)
            elements.setdefault(start, []).append(Start(markup))
            elements.setdefault(end, []).append(End(markup))
            if isinstance(markup, List):
                lis = markup['lis']
                for i, li in enumerate(lis):
                    start = li['start']
                    split_points.append(start)
                    li_end = lis[i + 1]['start'] if i < len(lis) - 1 else end
                    elements.setdefault(li_end, []).append(End(li))
                    elements.setdefault(start, []).append(Start(li))
        split_points = set(split_points)
        new_body = [[]]
        for i, letter in enumerate(body + '\x00'):
            if i in split_points:
                body_elements = []
                for elem in elements[i]:
                    body_elements.append(elem)
                new_body.append(body_elements)
                new_body.append([])
            new_body[-1].append(letter)
        new_body[-1] = new_body[-1][:-1]
        final = []
        for chunk in new_body:
            if not chunk:
                continue
            final.append(''.join(chunk) if isinstance(chunk[0], str) else chunk)
        return final

    def to_plain_text(self, body, markup_elem):
        chunks = self._split_first_level(body, markup_elem)
        final = []
        for chunk in chunks:
            if isinstance(chunk, str):
                final.append(chunk)
        return ''.join(final)

    def to_xhtml_im(self, body, markup_elem):
        chunks = self._split_first_level(body, markup_elem)
        final = []
        stack = []
        for chunk in chunks:
            if isinstance(chunk, str):
                chunk = (chunk.replace("&", '&amp;')
                              .replace('<', '&lt;')
                              .replace('>', '&gt;')
                              .replace('"', '&quot;')
                              .replace("'", '&apos;')
                              .replace('\n', '<br/>'))
                final.append(chunk)
                continue
            num_end = 0
            for elem in chunk:
                if isinstance(elem, End):
                    num_end += 1

            for i in range(num_end):
                stack_top = stack.pop()
                for elem in chunk:
                    if not isinstance(elem, End):
                        continue
                    elem = elem.elem
                    if elem is stack_top:
                        if isinstance(elem, Span):
                            final.append('</span>')
                        elif isinstance(elem, BlockCode):
                            final.append('</code></pre>')
                        elif isinstance(elem, List):
                            final.append('</ul>')
                        elif isinstance(elem, Li):
                            final.append('</li>')
                        elif isinstance(elem, BlockQuote):
                            final.append('</blockquote>')
                        break
                else:
                    assert False
            for elem in chunk:
                if not isinstance(elem, Start):
                    continue
                elem = elem.elem
                stack.append(elem)
                if isinstance(elem, Span):
                    style = []
                    for type_ in elem['types']:
                        if type_ == 'emphasis':
                            style.append('font-style: italic;')
                        if type_ == 'code':
                            style.append('font-family: monospace;')
                        if type_ == 'deleted':
                            style.append('text-decoration: line-through;')
                    final.append("<span style='%s'>" % ' '.join(style))
                elif isinstance(elem, BlockCode):
                    final.append('<pre><code>')
                elif isinstance(elem, List):
                    final.append('<ul>')
                elif isinstance(elem, Li):
                    final.append('<li>')
                elif isinstance(elem, BlockQuote):
                    final.append('<blockquote>')
        p = "<p xmlns='http://www.w3.org/1999/xhtml'>%s</p>" % ''.join(final)
        p2 = ET.fromstring(p)
        print('coucou', p, tostring(p2))
        xhtml_im = XHTML_IM()
        xhtml_im['body'] = p2
        return xhtml_im




# Slixmpp: The Slick XMPP Library
# Copyright (C) 2020 "Maxime “pep” Buquet <pep@bouah.net>"
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
from slixmpp import JID, Message, Presence
from slixmpp.plugins import BasePlugin
from slixmpp.xmlstream import register_stanza_plugin
from slixmpp.plugins.xep_0421 import stanza
from slixmpp.plugins.xep_0421.stanza import OccupantId


class XEP_0421(BasePlugin):
    '''XEP-0421: Anonymous unique occupant identifiers for MUCs'''

    name = 'xep_0421'
    description = 'XEP-0421: Anonymous unique occupant identifiers for MUCs'
    dependencies = {'xep_0030', 'xep_0045'}
    stanza = stanza
    namespace = stanza.NS

    def plugin_init(self) -> None:
        # XXX: This should be MucMessage. Someday..
        register_stanza_plugin(Message, OccupantId)
        register_stanza_plugin(Presence, OccupantId)

    async def has_feature(self, jid: JID, **discokwargs) -> bool:
        """
        Check if the remote JID supports Occupant-Id
        This method can take all the additional kwargs that can be
        provided to XEP-0030’s get_info.

        :param jid: JID of the remote entity.
        :returns: ``True`` if the server supports occupant-id.
        """
        info = await self.xmpp['xep_0030'].get_info(jid, **discokwargs)
        return self.namespace in info.get_features()




# Slixmpp: The Slick XMPP Library
# Copyright (C) 2011 Nathanael C. Fritz, Lance J.T. Stout
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
from slixmpp.stanza import Error
from slixmpp.xmlstream import ElementBase, ET


class LegacyError(ElementBase):

    """
    Older XMPP implementations used code based error messages, similar
    to HTTP response codes. Since then, error condition elements have
    been introduced. XEP-0086 provides a mapping between the new
    condition elements and a combination of error types and the older
    response codes.

    Also see <http://xmpp.org/extensions/xep-0086.html>.

    Example legacy error stanzas:
    ::

        <error xmlns="jabber:client" code="501" type="cancel">
          <feature-not-implemented
                xmlns="urn:ietf:params:xml:ns:xmpp-stanzas" />
        </error>

        <error code="402" type="auth">
          <payment-required
                xmlns="urn:ietf:params:xml:ns:xmpp-stanzas" />
        </error>

    :var error_map: A map of error conditions to error types and
                    code values.
    """

    name = 'legacy'
    namespace = Error.namespace
    plugin_attrib = name
    interfaces = {'condition'}
    overrides = ['set_condition']

    error_map = {'bad-request': ('modify', '400'),
                 'conflict': ('cancel', '409'),
                 'feature-not-implemented': ('cancel', '501'),
                 'forbidden': ('auth', '403'),
                 'gone': ('modify', '302'),
                 'internal-server-error': ('wait', '500'),
                 'item-not-found': ('cancel', '404'),
                 'jid-malformed': ('modify', '400'),
                 'not-acceptable': ('modify', '406'),
                 'not-allowed': ('cancel', '405'),
                 'not-authorized': ('auth', '401'),
                 'payment-required': ('auth', '402'),
                 'recipient-unavailable': ('wait', '404'),
                 'redirect': ('modify', '302'),
                 'registration-required': ('auth', '407'),
                 'remote-server-not-found': ('cancel', '404'),
                 'remote-server-timeout': ('wait', '504'),
                 'resource-constraint': ('wait', '500'),
                 'service-unavailable': ('cancel', '503'),
                 'subscription-required': ('auth', '407'),
                 'undefined-condition': (None, '500'),
                 'unexpected-request': ('wait', '400')}

    def setup(self, xml):
        """Don't create XML for the plugin."""
        self.xml = ET.Element('')

    def set_condition(self, value):
        """
        Set the error type and code based on the given error
        condition value.

        :param value: The new error condition.
        """
        self.parent().set_condition(value)

        error_data = self.error_map.get(value, None)
        if error_data is not None:
            if error_data[0] is not None:
                self.parent()['type'] = error_data[0]
            self.parent()['code'] = error_data[1]




# Slixmpp: The Slick XMPP Library
# Copyright (C) 2011 Nathanael C. Fritz, Dann Martens (TOMOTON).
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
import logging

from slixmpp import Iq
from slixmpp.xmlstream import ET, register_stanza_plugin
from slixmpp.xmlstream.handler import Callback
from slixmpp.xmlstream.matcher import MatchXPath
from slixmpp.plugins import BasePlugin
from slixmpp.plugins.xep_0009 import stanza
from slixmpp.plugins.xep_0009.stanza import RPCQuery, MethodCall, MethodResponse


log = logging.getLogger(__name__)


class XEP_0009(BasePlugin):

    name = 'xep_0009'
    description = 'XEP-0009: Jabber-RPC'
    dependencies = {'xep_0030'}
    stanza = stanza

    def plugin_init(self):
        register_stanza_plugin(Iq, RPCQuery)
        register_stanza_plugin(RPCQuery, MethodCall)
        register_stanza_plugin(RPCQuery, MethodResponse)

        self.xmpp.register_handler(
            Callback('RPC Call', MatchXPath('{%s}iq/{%s}query/{%s}methodCall' % (self.xmpp.default_ns, RPCQuery.namespace, RPCQuery.namespace)),
            self._handle_method_call)
        )
        self.xmpp.register_handler(
            Callback('RPC Call', MatchXPath('{%s}iq/{%s}query/{%s}methodResponse' % (self.xmpp.default_ns, RPCQuery.namespace, RPCQuery.namespace)),
            self._handle_method_response)
        )
        #self.activeCalls = []

        self.xmpp['xep_0030'].add_feature('jabber:iq:rpc')
        self.xmpp['xep_0030'].add_identity('automation','rpc')

    def make_iq_method_call(self, pto, pmethod, params):
        iq = self.xmpp.make_iq_set()
        iq['to'] = pto
        iq['from'] = self.xmpp.boundjid.full
        iq.enable('rpc_query')
        iq['rpc_query']['method_call']['method_name'] = pmethod
        iq['rpc_query']['method_call']['params'] = params
        return iq

    def make_iq_method_response(self, pid, pto, params):
        iq = self.xmpp.make_iq_result(pid)
        iq['to'] = pto
        iq['from'] = self.xmpp.boundjid.full
        iq.enable('rpc_query')
        iq['rpc_query']['method_response']['params'] = params
        return iq

    def make_iq_method_response_fault(self, pid, pto, params):
        iq = self.xmpp.make_iq_result(pid)
        iq['to'] = pto
        iq['from'] = self.xmpp.boundjid.full
        iq.enable('rpc_query')
        iq['rpc_query']['method_response']['params'] = None
        iq['rpc_query']['method_response']['fault'] = params
        return iq

#    def make_iq_method_error(self, pto, pid, pmethod, params, code, type, condition):
#        iq = self.xmpp.make_iq_error(pid)
#        iq.attrib['to'] = pto
#        iq.attrib['from'] = self.xmpp.boundjid.full
#        iq['error']['code'] = code
#        iq['error']['type'] = type
#        iq['error']['condition'] = condition
#        iq['rpc_query']['method_call']['method_name'] = pmethod
#        iq['rpc_query']['method_call']['params'] = params
#        return iq

    def _item_not_found(self, iq):
        payload = iq.get_payload()
        iq = iq.reply()
        iq.error().set_payload(payload)
        iq['error']['code'] = '404'
        iq['error']['type'] = 'cancel'
        iq['error']['condition'] = 'item-not-found'
        return iq

    def _undefined_condition(self, iq):
        payload = iq.get_payload()
        iq = iq.reply()
        iq.error().set_payload(payload)
        iq['error']['code'] = '500'
        iq['error']['type'] = 'cancel'
        iq['error']['condition'] = 'undefined-condition'
        return iq

    def _forbidden(self, iq):
        payload = iq.get_payload()
        iq = iq.reply()
        iq.error().set_payload(payload)
        iq['error']['code'] = '403'
        iq['error']['type'] = 'auth'
        iq['error']['condition'] = 'forbidden'
        return iq

    def _recipient_unavailable(self, iq):
        payload = iq.get_payload()
        iq = iq.reply()
        iq.error().set_payload(payload)
        iq['error']['code'] = '404'
        iq['error']['type'] = 'wait'
        iq['error']['condition'] = 'recipient-unavailable'
        return iq

    def _handle_method_call(self, iq):
        type = iq['type']
        if type == 'set':
            log.debug("Incoming Jabber-RPC call from %s", iq['from'])
            self.xmpp.event('jabber_rpc_method_call', iq)
        else:
            if type == 'error' and not iq['rpc_query'] == None:
                self.handle_error(iq)
            else:
                log.debug("Incoming Jabber-RPC error from %s", iq['from'])
                self.xmpp.event('jabber_rpc_error', iq)

    def _handle_method_response(self, iq):
        if iq['rpc_query']['method_response']['fault'] is not None:
            log.debug("Incoming Jabber-RPC fault from %s", iq['from'])
            #self._on_jabber_rpc_method_fault(iq)
            self.xmpp.event('jabber_rpc_method_fault', iq)
        else:
            log.debug("Incoming Jabber-RPC response from %s", iq['from'])
            self.xmpp.event('jabber_rpc_method_response', iq)

    def _send_fault(self, iq, fault_xml): #
        fault = self.make_iq_method_response_fault(iq['id'], iq['from'], fault_xml)
        fault.send()

    def _extract_method(self, stanza):
        xml = ET.fromstring("%s" % stanza)
        return xml.find("./methodCall/methodName").text



# Slixmpp: The Slick XMPP Library
# Copyright (C) 2025 Mathieu Pasquet
# This file is part of Slixmpp.
# See the file LICENSE for copying permissio
from slixmpp.plugins.base import register_plugin

from slixmpp.plugins.xep_0482 import stanza
from slixmpp.plugins.xep_0482.call_invites import XEP_0482


register_plugin(XEP_0482)

__all__ = ['stanza', 'XEP_0482']




# Slixmpp: The Slick XMPP Library
# Copyright (C) 2013 Nathanael C. Fritz, Lance J.T. Stout
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
from slixmpp.xmlstream import ElementBase


class FeatureNegotiation(ElementBase):

    name = 'feature'
    namespace = 'http://jabber.org/protocol/feature-neg'
    plugin_attrib = 'feature_neg'
    interfaces = set()




# slixmpp.xmlstream.matcher.id
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# Part of Slixmpp: The Slick XMPP Library
# :copyright: (c) 2011 Nathanael C. Fritz
# :license: MIT, see LICENSE for more details
from slixmpp.xmlstream.matcher.base import MatcherBase
from slixmpp.xmlstream.stanzabase import StanzaBase


class MatcherId(MatcherBase):

    """
    The ID matcher selects stanzas that have the same stanza 'id'
    interface value as the desired ID.
    """
    _criteria: str

    def match(self, xml: StanzaBase) -> bool:
        """Compare the given stanza's ``'id'`` attribute to the stored
        ``id`` value.

        :param xml: The :class:`~slixmpp.xmlstream.stanzabase.StanzaBase`
                    stanza to compare against.
        """
        return bool(xml.get_toplevel_attr("id", default='') == self._criteria)




# Slixmpp: The Slick XMPP Library
# Copyright (C) 2012 Nathanael C. Fritz, Lance J.T. Stout
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
import hashlib
import logging
from asyncio import Future

from slixmpp import JID
from slixmpp.stanza import Presence
from slixmpp.exceptions import XMPPError, IqTimeout, IqError
from slixmpp.xmlstream import register_stanza_plugin, ElementBase
from slixmpp.plugins.base import BasePlugin
from slixmpp.plugins.xep_0153 import stanza, VCardTempUpdate


log = logging.getLogger(__name__)


class XEP_0153(BasePlugin):

    name = 'xep_0153'
    description = 'XEP-0153: vCard-Based Avatars'
    dependencies = {'xep_0054'}
    stanza = stanza

    def plugin_init(self):
        self._hashes = {}

        register_stanza_plugin(Presence, VCardTempUpdate)

        self.xmpp.add_filter('out', self._update_presence)

        self.xmpp.add_event_handler('session_start', self._start)

        self.xmpp.add_event_handler('presence_available', self._recv_presence)
        self.xmpp.add_event_handler('presence_dnd', self._recv_presence)
        self.xmpp.add_event_handler('presence_xa', self._recv_presence)
        self.xmpp.add_event_handler('presence_chat', self._recv_presence)
        self.xmpp.add_event_handler('presence_away', self._recv_presence)

        self.api.register(self._set_hash, 'set_hash', default=True)
        self.api.register(self._get_hash, 'get_hash', default=True)
        self.api.register(self._reset_hash, 'reset_hash', default=True)

    def plugin_end(self):
        self.xmpp.del_filter('out', self._update_presence)
        self.xmpp.del_event_handler('session_start', self._start)
        self.xmpp.del_event_handler('session_end', self._end)
        self.xmpp.del_event_handler('presence_available', self._recv_presence)
        self.xmpp.del_event_handler('presence_dnd', self._recv_presence)
        self.xmpp.del_event_handler('presence_xa', self._recv_presence)
        self.xmpp.del_event_handler('presence_chat', self._recv_presence)
        self.xmpp.del_event_handler('presence_away', self._recv_presence)

    def set_avatar(self, jid: JID | None = None,
                   avatar: bytes | None = None,
                   mtype: str | None = None, **iqkwargs) -> Future:
        """Set a VCard avatar.

        :param jid: The JID to set the avatar for.
        :param avatar: Avatar content.
        :param mtype: Avatar file type (e.g. image/jpeg).
        """
        if jid is None:
            jid = self.xmpp.boundjid.bare

        async def get_and_set_avatar():
            timeout = iqkwargs.get('timeout', None)
            try:
                result = await self.xmpp['xep_0054'].get_vcard(
                    jid,
                    cached=False,
                    timeout=timeout
                )
            except IqTimeout:
                raise
            vcard = result['vcard_temp']
            vcard['PHOTO']['TYPE'] = mtype
            vcard['PHOTO']['BINVAL'] = avatar

            try:
                result = await self.xmpp['xep_0054'].publish_vcard(
                    jid=jid,
                    vcard=vcard,
                    **iqkwargs
                )
            except IqTimeout:
                raise
            await self.api['reset_hash'](jid)
            self.xmpp.roster[jid].send_last_presence()

        return self.xmpp.wrap(get_and_set_avatar())

    async def _start(self, event):
        try:
            vcard = await self.xmpp['xep_0054'].get_vcard(self.xmpp.boundjid.bare)
            data = vcard['vcard_temp']['PHOTO']['BINVAL']
            if not data:
                new_hash = ''
            else:
                new_hash = hashlib.sha1(data).hexdigest()
            await self.api['set_hash'](self.xmpp.boundjid, args=new_hash)
        except XMPPError:
            log.debug('Could not retrieve vCard for %s', self.xmpp.boundjid.bare)

    async def _update_presence(self, stanza: ElementBase) -> ElementBase:
        if not isinstance(stanza, Presence):
            return stanza

        if stanza['type'] not in ('available', 'dnd', 'chat', 'away', 'xa'):
            return stanza

        current_hash = await self.api['get_hash'](stanza['from'])
        stanza['vcard_temp_update']['photo'] = current_hash
        return stanza

    async def _recv_presence(self, pres: Presence):
        try:
            if pres.get_plugin('muc', check=True):
                # Don't process vCard avatars for MUC occupants
                # since they all share the same bare JID.
                return
        except:
            pass

        if not pres.match('presence/vcard_temp_update'):
            await self.api['set_hash'](pres['from'], args=None)
            return

        data = pres['vcard_temp_update']['photo']
        if data is None:
            return
        self.xmpp.event('vcard_avatar_update', pres)

    # =================================================================

    async def _reset_hash(self, jid: JID, node: str, ifrom: JID, args: dict):
        own_jid = (jid.bare == self.xmpp.boundjid.bare)
        if self.xmpp.is_component:
            own_jid = (jid.domain == self.xmpp.boundjid.domain)

        await self.api['set_hash'](jid, args=None)
        if own_jid:
            self.xmpp.roster[jid].send_last_presence()

        try:
            iq = await self.xmpp['xep_0054'].get_vcard(jid=jid.bare, ifrom=ifrom)
        except (IqError, IqTimeout):
            log.debug('Could not retrieve vCard for %s', jid)
            return
        try:
            data = iq['vcard_temp']['PHOTO']['BINVAL']
        except ValueError:
            log.debug('Invalid BINVAL in vCard’s PHOTO for %s:', jid, exc_info=True)
            data = None
        if not data:
            new_hash = ''
        else:
            new_hash = hashlib.sha1(data).hexdigest()

        await self.api['set_hash'](jid, args=new_hash)

    def _get_hash(self, jid: JID, node: str, ifrom: JID, args: dict):
        return self._hashes.get(jid.bare, None)

    def _set_hash(self, jid: JID, node: str, ifrom: JID, args: dict):
        self._hashes[jid.bare] = args




# Slixmpp: The Slick XMPP Library
# Copyright (C) 2012 Nathanael C. Fritz, Lance J.T. Stout
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
import logging

from slixmpp import Iq, Message
from slixmpp.plugins import BasePlugin
from slixmpp.xmlstream import register_stanza_plugin
from slixmpp.plugins.xep_0258 import stanza, SecurityLabel, Catalog


log = logging.getLogger(__name__)


class XEP_0258(BasePlugin):

    name = 'xep_0258'
    description = 'XEP-0258: Security Labels in XMPP'
    dependencies = {'xep_0030'}
    stanza = stanza

    def plugin_init(self):
        register_stanza_plugin(Message, SecurityLabel)
        register_stanza_plugin(Iq, Catalog)

    def plugin_end(self):
        self.xmpp['xep_0030'].del_feature(feature=SecurityLabel.namespace)

    def session_bind(self, jid):
        self.xmpp['xep_0030'].add_feature(SecurityLabel.namespace)

    def get_catalog(self, jid, ifrom=None,
                          callback=None, timeout=None):
        iq = self.xmpp.Iq()
        iq['to'] = jid
        iq['from'] = ifrom
        iq['type'] = 'get'
        iq.enable('security_label_catalog')
        return iq.send(callback=callback, timeout=timeout)



# Slixmpp: The Slick XMPP Library
# Copyright (C) 2012 Nathanael C. Fritz, Lance J.T. Stout
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
from slixmpp.plugins.base import register_plugin

from slixmpp.plugins.xep_0153.stanza import VCardTempUpdate
from slixmpp.plugins.xep_0153.vcard_avatar import XEP_0153


register_plugin(XEP_0153)

__all__ = ['VCardTempUpdate', 'XEP_0153']




# Slixmpp: The Slick XMPP Library
# Copyright (C) 2011 Nathanael C. Fritz, Lance J.T. Stout
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
from slixmpp import Message, Presence
from slixmpp.xmlstream import register_stanza_plugin
from slixmpp.plugins import BasePlugin
from slixmpp.plugins.xep_0131 import stanza
from slixmpp.plugins.xep_0131.stanza import Headers


class XEP_0131(BasePlugin):

    name = 'xep_0131'
    description = 'XEP-0131: Stanza Headers and Internet Metadata'
    dependencies = {'xep_0030'}
    stanza = stanza
    default_config = {
        'supported_headers': set()
    }

    def plugin_init(self):
        register_stanza_plugin(Message, Headers)
        register_stanza_plugin(Presence, Headers)

    def plugin_end(self):
        self.xmpp['xep_0030'].del_feature(feature=Headers.namespace)
        for header in self.supported_headers:
            self.xmpp['xep_0030'].del_feature(
                    feature='%s#%s' % (Headers.namespace, header))

    def session_bind(self, jid):
        self.xmpp['xep_0030'].add_feature(Headers.namespace)
        for header in self.supported_headers:
            self.xmpp['xep_0030'].add_feature('%s#%s' % (
                Headers.namespace,
                header))



# Slixmpp: The Slick XMPP Library
# Copyright (C) 2025 Mathieu Pasquet
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.

from slixmpp.plugins.base import register_plugin

from .muc_activity import XEP_0502

register_plugin(XEP_0502)

__all__ = ["XEP_0502"]




# Slixmpp: The Slick XMPP Library
# Copyright (C) 2017 Emmanuel Gil Peyrot
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
from slixmpp.xmlstream import ElementBase


class Hash(ElementBase):
    """
    Hash element, contains the name of the hash algorithm and the
    matching digest.

    .. code-block:: xml

        <hash xmlns='urn:xmpp:hashes:2' algo='sha-256'>2XarmwTlNxDAMkvymloX3S5+VbylNrJt/l5QyPa+YoU=</hash>

    """
    name = 'hash'
    namespace = 'urn:xmpp:hashes:2'
    plugin_attrib = 'hash'
    interfaces = {'algo', 'value'}

    allowed_algos = [
        'sha-1',
        'sha-256', 'sha-512',
        'sha3-256', 'sha3-512',
        'BLAKE2b256', 'BLAKE2b512',
    ]

    def set_algo(self, value):
        if value in self.allowed_algos:
            self._set_attr('algo', value)
        elif value in [None, '']:
            self._del_attr('algo')
        else:
            raise ValueError('Invalid algo: %s' % value)

    def get_value(self):
        return self.xml.text

    def set_value(self, value):
        self.xml.text = value

    def del_value(self):
        self.xml.text = ''




# Slixmpp: The Slick XMPP Library
# Implementation of xeps for Internet of Things
# http://wiki.xmpp.org/web/Tech_pages/IoT_systems
# Copyright (C) 2013 Sustainable Innovation, Joachim.lindborg@sust.se, bjorn.westrom@consoden.se
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
import datetime
import logging

class Device(object):
    """
    Example implementation of a device readout object.
    Is registered in the XEP_0323.register_node call
    The device object may be any custom implementation to support
    specific devices, but it must implement the functions:
          has_field
          request_fields
    """

    def __init__(self, nodeId, fields=None):
        if not fields:
            fields = {}

        self.nodeId = nodeId
        self.fields = fields # see fields described below
        # {'type':'numeric',
        #  'name':'myname',
        #  'value': 42,
        #  'unit':'Z'}]
        self.timestamp_data = {}
        self.momentary_data = {}
        self.momentary_timestamp = ""
        logging.debug("Device object started nodeId %s",nodeId)

    def has_field(self, field):
        """
        Returns true if the supplied field name exists in this device.

        Arguments:
            field      -- The field name
        """
        if field in self.fields.keys():
            return True
        return False

    def refresh(self, fields):
        """
        override method to do the refresh work
        refresh values from hardware or other
        """
        pass


    def request_fields(self, fields, flags, session, callback):
        """
        Starts a data readout. Verifies the requested fields,
        refreshes the data (if needed) and calls the callback
        with requested data.


        Arguments:
            fields   -- List of field names to readout
            flags    -- [optional] data classifier flags for the field, e.g. momentary
                        Formatted as a dictionary like { "flag name": "flag value" ... }
            session  -- Session id, only used in the callback as identifier
            callback -- Callback function to call when data is available.

                    The callback function must support the following arguments:

                session  -- Session id, as supplied in the request_fields call
                nodeId   -- Identifier for this device
                result   -- The current result status of the readout. Valid values are:
                               "error"  - Readout failed.
                               "fields" - Contains readout data.
                               "done"   - Indicates that the readout is complete. May contain
                                          readout data.
                timestamp_block -- [optional] Only applies when result != "error"
                               The readout data. Structured as a dictionary:
                  {
                    timestamp:     timestamp for this datablock,
                    fields:        list of field dictionary (one per readout field).
                      readout field dictionary format:
                      {
                        type:      The field type (numeric, boolean, dateTime, timeSpan, string, enum)
                        name:      The field name
                        value:     The field value
                        unit:      The unit of the field. Only applies to type numeric.
                        dataType:  The datatype of the field. Only applies to type enum.
                        flags:     [optional] data classifier flags for the field, e.g. momentary
                               Formatted as a dictionary like { "flag name": "flag value" ... }
                      }
                  }
                error_msg -- [optional] Only applies when result == "error".
                                Error details when a request failed.

        """
        logging.debug("request_fields called looking for fields %s",fields)
        if len(fields) > 0:
            # Check availability
            for f in fields:
                if f not in self.fields.keys():
                    self._send_reject(session, callback)
                    return False
        else:
            # Request all fields
            fields = self.fields.keys()


        # Refresh data from device
        # ...
        logging.debug("about to refresh device fields %s",fields)
        self.refresh(fields)

        if "momentary" in flags and flags['momentary'] == "true" or \
           "all" in flags and flags['all'] == "true":
            ts_block = {}
            timestamp = ""

            if len(self.momentary_timestamp) > 0:
                timestamp = self.momentary_timestamp
            else:
                timestamp = self._get_timestamp()

            field_block = []
            for f in self.momentary_data:
                if f in fields:
                    field_block.append({"name": f,
                                "type": self.fields[f]["type"],
                                "unit": self.fields[f]["unit"],
                                "dataType": self.fields[f]["dataType"],
                                "value": self.momentary_data[f]["value"],
                                "flags": self.momentary_data[f]["flags"]})
            ts_block["timestamp"] = timestamp
            ts_block["fields"] = field_block

            callback(session, result="done", nodeId=self.nodeId, timestamp_block=ts_block)
            return

        from_flag = self._datetime_flag_parser(flags, 'from')
        to_flag = self._datetime_flag_parser(flags, 'to')

        for ts in sorted(self.timestamp_data.keys()):
            tsdt = datetime.datetime.strptime(ts, "%Y-%m-%dT%H:%M:%S")
            if from_flag is not None:
                if tsdt < from_flag:
                    #print (str(tsdt) + " < " + str(from_flag))
                    continue
            if to_flag is not None:
                if tsdt > to_flag:
                    #print (str(tsdt) + " > " + str(to_flag))
                    continue

            ts_block = {}
            field_block = []

            for f in self.timestamp_data[ts]:
                if f in fields:
                    field_block.append({"name": f,
                                "type": self.fields[f]["type"],
                                "unit": self.fields[f]["unit"],
                                "dataType": self.fields[f]["dataType"],
                                "value": self.timestamp_data[ts][f]["value"],
                                "flags": self.timestamp_data[ts][f]["flags"]})

            ts_block["timestamp"] = ts
            ts_block["fields"] = field_block
            callback(session, result="fields", nodeId=self.nodeId, timestamp_block=ts_block)
        callback(session, result="done", nodeId=self.nodeId, timestamp_block=None)

    def _datetime_flag_parser(self, flags, flagname):
        if flagname not in flags:
            return None

        dt = None
        try:
            dt = datetime.datetime.strptime(flags[flagname], "%Y-%m-%dT%H:%M:%S")
        except ValueError:
            # Badly formatted datetime, ignore it
            pass
        return dt


    def _get_timestamp(self):
        """
        Generates a properly formatted timestamp of current time
        """
        return datetime.datetime.now().replace(microsecond=0).isoformat()

    def _send_reject(self, session, callback):
        """
        Sends a reject to the caller

        Arguments:
            session  -- Session id, see definition in request_fields function
            callback -- Callback function, see definition in request_fields function
        """
        callback(session, result="error", nodeId=self.nodeId, timestamp_block=None, error_msg="Reject")

    def _add_field(self, name, typename, unit=None, dataType=None):
        """
        Adds a field to the device

        Arguments:
            name     -- Name of the field
            typename -- Type of the field (numeric, boolean, dateTime, timeSpan, string, enum)
            unit     -- [optional] only applies to "numeric". Unit for the field.
            dataType -- [optional] only applies to "enum". Datatype for the field.
        """
        self.fields[name] = {"type": typename, "unit": unit, "dataType": dataType}

    def _add_field_timestamp_data(self, name, timestamp, value, flags=None):
        """
        Adds timestamped data to a field

        Arguments:
            name      -- Name of the field
            timestamp -- Timestamp for the data (string)
            value     -- Field value at the timestamp
            flags     -- [optional] data classifier flags for the field, e.g. momentary
                         Formatted as a dictionary like { "flag name": "flag value" ... }
        """
        if name not in self.fields.keys():
            return False
        if timestamp not in self.timestamp_data:
            self.timestamp_data[timestamp] = {}

        self.timestamp_data[timestamp][name] = {"value": value, "flags": flags}
        return True

    def _add_field_momentary_data(self, name, value, flags=None):
        """
        Sets momentary data to a field

        Arguments:
            name      -- Name of the field
            value     -- Field value at the timestamp
            flags     -- [optional] data classifier flags for the field, e.g. momentary
                         Formatted as a dictionary like { "flag name": "flag value" ... }
        """
        if name not in self.fields:
            return False
        if flags is None:
            flags = {}

        flags["momentary"] = "true"
        self.momentary_data[name] = {"value": value, "flags": flags}
        return True

    def _set_momentary_timestamp(self, timestamp):
        """
        This function is only for unit testing to produce predictable results.
        """
        self.momentary_timestamp = timestamp



# slixmpp: The Slick XMPP Library
# Copyright (C) 2016 Emmanuel Gil Peyrot
# This file is part of slixmpp.
# See the file LICENSE for copying permission.
from slixmpp.plugins.base import register_plugin

from slixmpp.plugins.xep_0333.stanza import Markable, Received, Displayed, Acknowledged
from slixmpp.plugins.xep_0333.markers import XEP_0333

register_plugin(XEP_0333)

__all__ = ['Markable', 'Received', 'Displayed', 'Acknowledged', 'XEP_0333']




# Slixmpp: The Slick XMPP Library
# Copyright (C) 2010  Nathanael C. Fritz
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
import logging

from slixmpp import InvalidJID
from slixmpp.stanza import Iq
from slixmpp.xmlstream import JID
from slixmpp.xmlstream import ET, ElementBase, register_stanza_plugin


class Roster(ElementBase):

    """
    Example roster stanzas:
        <iq type="set">
          <query xmlns="jabber:iq:roster">
            <item jid="user@example.com" subscription="both" name="User">
              <group>Friends</group>
            </item>
          </query>
        </iq>

    Stanza Interface:
        items -- A dictionary of roster entries contained
                 in the stanza.

    Methods:
        get_items -- Return a dictionary of roster entries.
        set_items -- Add <item> elements.
        del_items -- Remove all <item> elements.
    """

    namespace = 'jabber:iq:roster'
    name = 'query'
    plugin_attrib = 'roster'
    interfaces = {'items', 'ver'}

    def get_ver(self):
        """
        Ensure handling an empty ver attribute property.

        The ver attribute is special in that the presence of the
        attribute with an empty value is important for bootstrapping
        roster versioning.
        """
        return self.xml.attrib.get('ver', None)

    def set_ver(self, ver):
        """
        Ensure handling an empty ver attribute property.

        The ver attribute is special in that the presence of the
        attribute with an empty value is important for bootstrapping
        roster versioning.
        """
        if ver is not None:
            self.xml.attrib['ver'] = ver
        else:
            del self.xml.attrib['ver']

    def set_items(self, items):
        """
        Set the roster entries in the <roster> stanza.

        Uses a dictionary using JIDs as keys, where each entry is itself
        a dictionary that contains:
            name         -- An alias or nickname for the JID.
            subscription -- The subscription type. Can be one of 'to',
                            'from', 'both', 'none', or 'remove'.
            groups       -- A list of group names to which the JID
                            has been assigned.

        Arguments:
            items -- A dictionary of roster entries.
        """
        self.del_items()
        for jid in items:
            item = RosterItem()
            item.values = items[jid]
            item['jid'] = jid
            self.append(item)
        return self

    def get_items(self):
        """
        Return a dictionary of roster entries.

        Each item is keyed using its JID, and contains:
            name         -- An assigned alias or nickname for the JID.
            subscription -- The subscription type. Can be one of 'to',
                            'from', 'both', 'none', or 'remove'.
            groups       -- A list of group names to which the JID has
                            been assigned.
        """
        items = {}
        for item in self['substanzas']:
            if isinstance(item, RosterItem):
                try:
                    items[item['jid']] = item.values
                except InvalidJID:
                    log.warning('Invalid JID in roster: %s', item)
                    continue
                # Remove extra JID reference to keep everything
                # backward compatible
                del items[item['jid']]['jid']
                del items[item['jid']]['lang']
        return items

    def del_items(self):
        """
        Remove all <item> elements from the roster stanza.
        """
        for item in self['substanzas']:
            if isinstance(item, RosterItem):
                self.xml.remove(item.xml)


class RosterItem(ElementBase):
    namespace = 'jabber:iq:roster'
    name = 'item'
    plugin_attrib = 'item'
    interfaces = {'jid', 'name', 'subscription', 'ask', 'approved', 'groups'}

    def get_jid(self):
        return JID(self._get_attr('jid', ''))

    def set_jid(self, jid):
        self._set_attr('jid', str(jid))

    def get_groups(self):
        groups = []
        for group in self.xml.findall('{%s}group' % self.namespace):
            if group.text:
                groups.append(group.text)
            else:
                groups.append('')
        return groups

    def set_groups(self, values):
        self.del_groups()
        for group in values:
            group_xml = ET.Element('{%s}group' % self.namespace)
            group_xml.text = group
            self.xml.append(group_xml)

    def del_groups(self):
        for group in self.xml.findall('{%s}group' % self.namespace):
            self.xml.remove(group)


register_stanza_plugin(Iq, Roster)
register_stanza_plugin(Roster, RosterItem, iterable=True)

log = logging.getLogger(__name__)



# Slixmpp: The Slick XMPP Library
# Copyright (C) 2021 Mathieu Pasquet
# This file is part of Slixmpp.
# See the file LICENSE for copying permission
import logging

from asyncio import Future

from slixmpp import JID
from slixmpp.types import MAMDefault
from slixmpp.stanza import Iq
from slixmpp.xmlstream import register_stanza_plugin
from slixmpp.plugins import BasePlugin
from slixmpp.plugins.xep_0441 import stanza


log = logging.getLogger(__name__)


class XEP_0441(BasePlugin):

    """
    XEP-0441: Message Archive Management Preferences
    """

    name = 'xep_0441'
    description = 'XEP-0441: Message Archive Management Preferences'
    dependencies = {'xep_0313'}
    stanza = stanza

    def plugin_init(self):
        register_stanza_plugin(Iq, stanza.Preferences)

    async def get_preferences(self, **iqkwargs
                              ) -> tuple[MAMDefault, list[JID], list[JID]]:
        """Get the current MAM preferences.

        Takes all parameters from ``iq.send()``.

        :returns: A tuple of MAM preferences with (default, always, never)
        """
        ifrom = iqkwargs.pop('ifrom', None)
        ito = iqkwargs.pop('ito', None)
        iq = self.xmpp.make_iq_get(ito=ito, ifrom=ifrom)
        iq['type'] = 'get'
        query_id = iq['id']
        iq['mam_prefs']['query_id'] = query_id
        result = await iq.send(**iqkwargs)
        return (
            result['mam_prefs']['default'],
            result['mam_prefs']['always'],
            result['mam_prefs']['never']
        )

    def set_preferences(self, default: MAMDefault | None = 'roster',
                        always: list[JID] | None = None,
                        never: list[JID] | None = None, *,
                        ito: JID | None = None, ifrom: JID | None = None,
                        **iqkwargs) -> Future[Iq]:
        """Set MAM Preferences.

        The server answer MAY contain different items.

        :param default: Default behavior (one of 'always', 'never', 'roster').
        :param always: List of JIDs whose messages will always be stored.
        :param never: List of JIDs whose messages will never be stored.
        """
        iq = self.xmpp.make_iq_set(ito=ito, ifrom=ifrom)
        iq['mam_prefs']['default'] = default
        iq['mam_prefs']['always'] = always
        iq['mam_prefs']['never'] = never
        return iq.send(**iqkwargs)




# Slixmpp: The Slick XMPP Library
# Copyright (C) 2021 Mathieu Pasquet <mathieui@mathieui.net>
# This file is part of Slixmpp.
# See the file LICENSE for copying permissio
from slixmpp.stanza import Message
from slixmpp.xmlstream import (
    register_stanza_plugin,
    ElementBase,
)


NS = 'urn:xmpp:spoiler:0'


class Spoiler(ElementBase):
    namespace = NS
    name = 'spoiler'
    plugin_attrib = 'spoiler'


def register_plugins():
    register_stanza_plugin(Message, Spoiler)




# Slixmpp: The Slick XMPP Library
# Copyright (C) 2012 Nathanael C. Fritz, Lance J.T. Stout
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.

from slixmpp.stanza import Message, Presence
from slixmpp.xmlstream import register_stanza_plugin
from slixmpp.plugins import BasePlugin
from slixmpp.plugins.xep_0091 import stanza


class XEP_0091(BasePlugin):

    """
    XEP-0091: Legacy Delayed Delivery
    """

    name = 'xep_0091'
    description = 'XEP-0091: Legacy Delayed Delivery'
    dependencies = set()
    stanza = stanza

    def plugin_init(self):
        register_stanza_plugin(Message, stanza.LegacyDelay)
        register_stanza_plugin(Presence, stanza.LegacyDelay)



from slixmpp.plugins.base import register_plugin
from slixmpp.plugins.xep_0122.stanza import FormValidation
from slixmpp.plugins.xep_0122.data_validation import XEP_0122


register_plugin(XEP_0122)

__all__ = ['FormValidation', 'XEP_0122']




# Slixmpp: The Slick XMPP Library
# Copyright (C) 2011 Nathanael C. Fritz, Dann Martens (TOMOTON).
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
from slixmpp.plugins.base import register_plugin

from slixmpp.plugins.xep_0009 import stanza
from slixmpp.plugins.xep_0009.rpc import XEP_0009
from slixmpp.plugins.xep_0009.stanza import RPCQuery, MethodCall, MethodResponse


register_plugin(XEP_0009)

__all__ = ['XEP_0009', 'RPCQuery', 'MethodCall', 'MethodResponse', 'stanza']




# Slixmpp: The Slick XMPP Library
# Copyright (C) 2012 Nathanael C. Fritz, Lance J.T. Stout
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
import logging

from slixmpp import Iq
from slixmpp.plugins import BasePlugin
from slixmpp.xmlstream import register_stanza_plugin
from slixmpp.plugins.xep_0186 import Visible, Invisible


log = logging.getLogger(__name__)


class XEP_0186(BasePlugin):

    name = 'xep_0186'
    description = 'XEP-0186: Invisible Command'
    dependencies = {'xep_0030'}

    def plugin_init(self):
        register_stanza_plugin(Iq, Visible)
        register_stanza_plugin(Iq, Invisible)

    def set_invisible(self, ifrom=None, callback=None,
                            timeout=None):
        iq = self.xmpp.Iq()
        iq['type'] = 'set'
        iq['from'] = ifrom
        iq.enable('invisible')
        return iq.send(callback=callback, timeout=timeout)

    def set_visible(self, ifrom=None, callback=None,
                          timeout=None):
        iq = self.xmpp.Iq()
        iq['type'] = 'set'
        iq['from'] = ifrom
        iq.enable('visible')
        return iq.send(callback=callback, timeout=timeout)



import logging

from slixmpp import Message, Iq, Presence, JID
from slixmpp.plugins import BasePlugin


class XEP_0100(BasePlugin):

    """
    XEP-0100: Gateway interaction

    Does not cover the deprecated Agent Information and 'jabber:iq:gateway' protocols

    Events registered by this plugin:

    - legacy_login: Jabber user got online or just registered
    - legacy_logout: Jabber user got offline or just unregistered
    - legacy_presence_unavailable: Jabber user sent an unavailable presence to a legacy contact
    - gateway_message: Jabber user sent a direct message to the gateway component
    - legacy_message: Jabber user sent a message to the legacy network


    Plugin Parameters:

    - `component_name`: (str) Name of the entity
    - `type`: (str) Type of the gateway identity. Should be the name of the legacy service
    - `needs_registration`: (bool) If set to True, messages received from unregistered users will
      not be transmitted to the legacy service

    API:

    - legacy_contact_add(jid, node, ifrom: JID, args: JID): Add contact on the legacy service.
      Should raise LegacyError if anything goes wrong in the process.
      `ifrom` is the gateway user's JID and `args` is the legacy contact's JID.
    - legacy_contact_remove(jid, node, ifrom: JID, args: JID): Remove a contact.

    """

    name = "xep_0100"
    description = "XEP-0100: Gateway interaction"
    dependencies = {
        "xep_0030",  # Service discovery
        "xep_0077",  # In band registration
    }

    default_config = {
        "component_name": "SliXMPP gateway",
        "type": "xmpp",
        "needs_registration": True,
    }

    def plugin_init(self):
        if not self.xmpp.is_component:
            log.error("Only components can be gateways, aborting plugin load")
            return

        self.xmpp["xep_0030"].add_identity(
            name=self.component_name, category="gateway", itype=self.type
        )

        self.api.register(self._legacy_contact_remove, "legacy_contact_remove")
        self.api.register(self._legacy_contact_add, "legacy_contact_add")

        # Without that BaseXMPP sends unsub/unavailable on sub requests and we don't want that
        self.xmpp.client_roster.auto_authorize = True
        self.xmpp.client_roster.auto_subscribe = False

        self.xmpp.add_event_handler("user_register", self.on_user_register)
        self.xmpp.add_event_handler("user_unregister", self.on_user_unregister)
        self.xmpp.add_event_handler("presence_available", self.on_presence_available)
        self.xmpp.add_event_handler(
            "presence_unavailable", self.on_presence_unavailable
        )
        self.xmpp.add_event_handler("presence_subscribe", self.on_presence_subscribe)
        self.xmpp.add_event_handler(
            "presence_unsubscribe", self.on_presence_unsubscribe
        )
        self.xmpp.add_event_handler("message", self.on_message)

    def plugin_end(self):
        if not self.xmpp.is_component:
            return

        self.xmpp.del_event_handler("user_register", self.on_user_register)
        self.xmpp.del_event_handler("user_unregister", self.on_user_unregister)
        self.xmpp.del_event_handler("presence_available", self.on_presence_available)
        self.xmpp.del_event_handler(
            "presence_unavailable", self.on_presence_unavailable
        )
        self.xmpp.del_event_handler("presence_subscribe", self.on_presence_subscribe)
        self.xmpp.del_event_handler("message", self.on_message)
        self.xmpp.del_event_handler(
            "presence_unsubscribe", self.on_presence_unsubscribe
        )

    async def get_user(self, stanza):
        return await self.xmpp["xep_0077"].api["user_get"](None, None, None, stanza)

    def send_presence(self, pto, ptype=None, pstatus=None, pfrom=None):
        self.xmpp.send_presence(
            pfrom=self.xmpp.boundjid.bare,
            ptype=ptype,
            pto=pto,
            pstatus=pstatus,
        )

    async def on_user_register(self, iq: Iq):
        user_jid = iq["from"]
        user = await self.get_user(iq)
        if user is None:  # This should not happen
            log.warning(f"{user_jid} has registered but cannot find them in user store")
        else:
            log.debug(f"Sending subscription request to {user_jid}")
            self.xmpp.client_roster.subscribe(user_jid)

    def on_user_unregister(self, iq: Iq):
        user_jid = iq["from"]
        log.debug(f"Sending subscription request to {user_jid}")
        self.xmpp.event("legacy_logout", iq)
        self.xmpp.client_roster.unsubscribe(iq["from"])
        self.xmpp.client_roster.remove(iq["from"])
        log.debug(f"roster: {self.xmpp.client_roster}")

    async def on_presence_available(self, presence: Presence):
        user_jid = presence["from"]
        user = await self.get_user(presence)
        if user is None:
            log.warning(
                f"{user_jid} has gotten online but cannot find them in user store"
            )
        else:
            self.xmpp.event("legacy_login", presence)
            log.debug(f"roster: {self.xmpp.client_roster}")
            self.send_presence(pto=user_jid.bare, ptype="available")

    async def on_presence_unavailable(self, presence: Presence):
        user_jid = presence["from"]
        user = await self.get_user(presence)
        if user is None:  # This should not happen
            log.warning(
                f"{user_jid} has gotten offline but but cannot find them in user store"
            )
            return

        if presence["to"] == self.xmpp.boundjid.bare:
            self.xmpp.event("legacy_logout", presence)
            self.send_presence(pto=user_jid, ptype="unavailable")
        else:
            self.xmpp.event("legacy_presence_unavailable", presence)

    async def _legacy_contact_add(self, jid, node, ifrom, contact_jid: JID):
        pass

    async def on_presence_subscribe(self, presence: Presence):
        user_jid = presence["from"]
        user = await self.get_user(presence)
        if user is None and self.needs_registration:
            return

        if presence["to"] == self.xmpp.boundjid.bare:
            return

        try:
            await self.api["legacy_contact_add"](
                ifrom=user_jid,
                args=presence["to"],
            )
        except LegacyError:
            self.xmpp.send_presence(
                pfrom=presence["to"],
                ptype="unsubscribed",
                pto=user_jid,
            )
            return
        self.xmpp.send_presence(
            pfrom=presence["to"],
            ptype="subscribed",
            pto=user_jid,
        )
        self.xmpp.send_presence(
            pfrom=presence["to"],
            pto=user_jid,
        )
        self.xmpp.send_presence(
            pfrom=presence["to"],
            ptype="subscribe",
            pto=user_jid,
        )  # TODO: handle resulting subscribed presences

    async def on_presence_unsubscribe(self, presence: Presence):
        if presence["to"] == self.xmpp.boundjid.bare:
            # should we trigger unregistering here?
            return

        user_jid = presence["from"]
        user = await self.get_user(presence)
        if user is None:
            log.debug("Received remove subscription from unregistered user")
            if self.needs_registration:
                return

        await self.api["legacy_contact_remove"](ifrom=user_jid, args=presence["to"])

        for ptype in "unsubscribe", "unsubscribed", "unavailable":
            self.xmpp.send_presence(
                pfrom=presence["to"],
                ptype=ptype,
                pto=user_jid,
            )

    async def _legacy_contact_remove(self, jid, node, ifrom, contact_jid: JID):
        pass

    async def on_message(self, msg: Message):
        if msg["type"] == "groupchat":
            return  # groupchat messages are out of scope of XEP-0100

        if msg["to"] == self.xmpp.boundjid.bare:
            # It may be useful to exchange direct messages with the component
            self.xmpp.event("gateway_message", msg)
            return

        if self.needs_registration and await self.get_user(msg) is None:
            return

        self.xmpp.event("legacy_message", msg)

    def transform_legacy_message(
        self,
        jabber_user_jid: JID | str,
        legacy_contact_id: str,
        body: str,
        mtype: str | None = None,
    ):
        """
        Transform a legacy message to an XMPP message
        """
        # Should escaping legacy IDs to valid JID local parts be handled here?
        # Maybe by internal API stuff?
        self.xmpp.send_message(
            mfrom=JID(f"{legacy_contact_id}@{self.xmpp.boundjid.bare}"),
            mto=JID(jabber_user_jid).bare,
            mbody=body,
            mtype=mtype,
        )


class LegacyError(Exception):
    pass


log = logging.getLogger(__name__)




# Slixmpp: The Slick XMPP Library
# Copyright (C) 2011 Nathanael C. Fritz, Lance J.T. Stout
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.

from slixmpp.stanza import Message, Presence
from slixmpp.xmlstream import register_stanza_plugin
from slixmpp.plugins import BasePlugin
from slixmpp.plugins.xep_0203 import stanza


class XEP_0203(BasePlugin):

    """
    XEP-0203: Delayed Delivery

    XMPP stanzas are sometimes withheld for delivery due to the recipient
    being offline, or are resent in order to establish recent history as
    is the case with MUCS. In any case, it is important to know when the
    stanza was originally sent, not just when it was last received.

    Also see <http://www.xmpp.org/extensions/xep-0203.html>.
    """

    name = 'xep_0203'
    description = 'XEP-0203: Delayed Delivery'
    dependencies = set()
    stanza = stanza

    def plugin_init(self):
        """Start the XEP-0203 plugin."""
        register_stanza_plugin(Message, stanza.Delay)
        register_stanza_plugin(Presence, stanza.Delay)




# Slixmpp: The Slick XMPP Library
# Copyright (C) 2011 Nathanael C. Fritz, Lance J.T. Stout
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
from slixmpp.stanza import Error
from slixmpp.xmlstream import register_stanza_plugin
from slixmpp.plugins import BasePlugin
from slixmpp.plugins.xep_0086 import stanza, LegacyError


class XEP_0086(BasePlugin):

    """
    XEP-0086: Error Condition Mappings

    Older XMPP implementations used code based error messages, similar
    to HTTP response codes. Since then, error condition elements have
    been introduced. XEP-0086 provides a mapping between the new
    condition elements and a combination of error types and the older
    response codes.

    Also see <http://xmpp.org/extensions/xep-0086.html>.

    Configuration Values:
    ::

        override -- Indicates if applying legacy error codes should
                    be done automatically. Defaults to True.
                    If False, then inserting legacy error codes can
                    be done using:
                        iq['error']['legacy']['condition'] = ...
    """

    name = 'xep_0086'
    description = 'XEP-0086: Error Condition Mappings'
    dependencies = set()
    stanza = stanza
    default_config = {
        'override': True
    }

    def plugin_init(self):
        register_stanza_plugin(Error, LegacyError,
                               overrides=self.override)



# slixmpp: The Slick XMPP Library
# Copyright (C) 2018 Emmanuel Gil Peyrot
# This file is part of slixmpp.
# See the file LICENSE for copying permission.
from slixmpp.plugins.base import register_plugin

from slixmpp.plugins.xep_0363.stanza import Request, Slot, Put, Get, Header
from slixmpp.plugins.xep_0363.http_upload import (
    XEP_0363,
    UploadServiceNotFound,
    FileTooBig,
    HTTPError,
    FileUploadError,
)

register_plugin(XEP_0363)

__all__ = ['Request', 'Slot', 'Put', 'Get', 'Header', 'XEP_0363',
           'UploadServiceNotFound', 'FileTooBig', 'HTTPError',
           'FileUploadError']




# Slixmpp: The Slick XMPP Library
# Copyright (C) 2013 Nathanael C. Fritz, Lance J.T. Stout
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
import datetime as dt

from slixmpp.xmlstream import ElementBase
from slixmpp.plugins import xep_0082


class Idle(ElementBase):
    name = 'idle'
    namespace = 'urn:xmpp:idle:1'
    plugin_attrib = 'idle'
    interfaces = {'since'}

    def get_since(self):
        timestamp = self._get_attr('since')
        return xep_0082.parse(timestamp)

    def set_since(self, value):
        if isinstance(value, dt.datetime):
            value = xep_0082.format_datetime(value)
        self._set_attr('since', value)



# Slixmpp: The Slick XMPP Library
# Copyright (C) 2011  Nathanael C. Fritz
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
import logging
from typing import TYPE_CHECKING, ClassVar, NamedTuple

from slixmpp.plugins import BasePlugin
from slixmpp.stanza import StreamFeatures
from slixmpp.xmlstream import register_stanza_plugin

from . import stanza

if TYPE_CHECKING:
    from slixmpp.clientxmpp import ClientXMPP


class FeatureLimits(BasePlugin):
    xmpp: "ClientXMPP"
    name = "feature_limits"
    description = "XEP-0478: Stream Limits Advertisement"
    dependencies: ClassVar[set[str]] = set()
    stanza = stanza

    def plugin_init(self):
        if self.xmpp.is_component:
            raise RuntimeError("Stream limits does not work with XEP-0114")

        self.xmpp.register_feature(
            "limits", self._handle_limits, restart=False, order=10000
        )

        register_stanza_plugin(StreamFeatures, stanza.Limits)

    async def _handle_limits(self, features: StreamFeatures):
        self.xmpp.limits = Limits(
            _attempt_int(features["limits"]["max-bytes"]),
            _attempt_int(features["limits"]["idle-seconds"]),
        )


class Limits(NamedTuple):
    max_bytes: int | None = None
    idle_seconds: int | None = None


def _attempt_int(val: str) -> int | None:
    if not val:
        return None
    try:
        return int(val)
    except ValueError:
        log.warning("Received an invalid stream limit value: %s", val)
        return None


log = logging.getLogger(__name__)




# Slixmpp: The Slick XMPP Library
# Copyright (C) 2012 Nathanael C. Fritz, Lance J.T. Stout
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
import logging

from asyncio import Future

from slixmpp.stanza import Iq
from slixmpp.plugins import BasePlugin
from slixmpp.xmlstream.handler import Callback
from slixmpp.xmlstream.matcher import StanzaPath
from slixmpp.xmlstream import register_stanza_plugin, JID
from slixmpp.plugins.xep_0191 import stanza, Block, Unblock, BlockList, BlockItem


log = logging.getLogger(__name__)

BlockedJIDs = (
    JID |
    set[JID] |
    list[JID]
)


class XEP_0191(BasePlugin):

    name = 'xep_0191'
    description = 'XEP-0191: Blocking Command'
    dependencies = {'xep_0030'}
    stanza = stanza

    def plugin_init(self):
        register_stanza_plugin(Iq, BlockList)
        register_stanza_plugin(Iq, Block)
        register_stanza_plugin(Iq, Unblock)
        register_stanza_plugin(BlockList, BlockItem, iterable=True)
        register_stanza_plugin(Block, BlockItem, iterable=True)
        register_stanza_plugin(Unblock, BlockItem, iterable=True)

        self.xmpp.register_handler(
                Callback('Blocked Contact',
                    StanzaPath('iq@type=set/block'),
                    self._handle_blocked))

        self.xmpp.register_handler(
                Callback('Unblocked Contact',
                    StanzaPath('iq@type=set/unblock'),
                    self._handle_unblocked))

    def plugin_end(self):
        self.xmpp.remove_handler('Blocked Contact')
        self.xmpp.remove_handler('Unblocked Contact')

    def get_blocked(self, ifrom: JID | None = None, **iqkwargs) -> Future:
        """Get the iq containing the blocklist."""
        iq = self.xmpp.make_iq_get(ifrom=ifrom)
        iq.enable('blocklist')
        return iq.send(**iqkwargs)

    async def get_blocked_jids(self, ifrom: JID | None = None, **iqkwargs) -> set[JID]:
        """Get the list of blocked JIDs."""
        iq = self.xmpp.make_iq_get(ifrom=ifrom)
        iq.enable('blocklist')
        result = await iq.send(**iqkwargs)
        return {JID(item['jid']) for item in result['blocklist']}

    def block(self, jids: BlockedJIDs,
              ifrom: JID | None = None, **iqkwargs) -> Future:
        """Block a JID or a list of JIDs.

        :param jids: JID(s) to block.
        """
        iq = self.xmpp.make_iq_set(ifrom=ifrom)
        if not isinstance(jids, (set, list)):
            jids = [jids]

        for jid in jids:
            item = BlockItem(parent=iq['block'])
            item['jid'] = jid

        return iq.send(**iqkwargs)

    def unblock(self, jids: BlockedJIDs, ifrom: JID | None = None, **iqkwargs) -> Future:
        """Unblock a JID or a list of JIDs.

        :param jids: JID(s) to unblock.
        """
        if jids is None:
            raise ValueError("jids cannot be empty.")
        iq = self.xmpp.make_iq_set(ifrom=ifrom)

        if not isinstance(jids, (set, list)):
            jids = [jids]

        for jid in jids:
            item = BlockItem(parent=iq['unblock'])
            item['jid'] = jid

        return iq.send(**iqkwargs)

    def _handle_blocked(self, iq):
        self.xmpp.event('blocked', iq)

    def _handle_unblocked(self, iq):
        self.xmpp.event('unblocked', iq)




# Slixmpp: The Slick XMPP Library
# Copyright (C) 2012 Nathanael C. Fritz, Lance J.T. Stout
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
from slixmpp.xmlstream import JID, ElementBase, register_stanza_plugin


class Addresses(ElementBase):

    name = 'addresses'
    namespace = 'http://jabber.org/protocol/address'
    plugin_attrib = 'addresses'
    interfaces = set()

    def add_address(self, atype='to', jid='', node='', uri='',
                          desc='', delivered=False):
        addr = Address(parent=self)
        addr['type'] = atype
        addr['jid'] = jid
        addr['node'] = node
        addr['uri'] = uri
        addr['desc'] = desc
        addr['delivered'] = delivered

        return addr

    # Additional methods for manipulating sets of addresses
    # based on type are generated below.


class Address(ElementBase):

    name = 'address'
    namespace = 'http://jabber.org/protocol/address'
    plugin_attrib = 'address'
    interfaces = {'type', 'jid', 'node', 'uri', 'desc', 'delivered'}

    address_types = {'bcc', 'cc', 'noreply', 'replyroom', 'replyto', 'to'}

    def get_jid(self):
        return JID(self._get_attr('jid'))

    def set_jid(self, value):
        self._set_attr('jid', str(value))

    def get_delivered(self):
        value = self._get_attr('delivered', False)
        return value and value.lower() in ('true', '1')

    def set_delivered(self, delivered):
        if delivered:
            self._set_attr('delivered', 'true')
        else:
            del self['delivered']

    def set_uri(self, uri):
        if uri:
            del self['jid']
            del self['node']
            self._set_attr('uri', uri)
        else:
            self._del_attr('uri')


# =====================================================================
# Auto-generate address type filters for the Addresses class.

def _addr_filter(atype):
    def _type_filter(addr):
        if isinstance(addr, Address):
            if atype == 'all' or addr['type'] == atype:
                return True
        return False
    return _type_filter


def _build_methods(atype):

    def get_multi(self):
        return list(filter(_addr_filter(atype), self))

    def set_multi(self, value):
        del self[atype]
        for addr in value:

            # Support assigning dictionary versions of addresses
            # instead of full Address objects.
            if not isinstance(addr, Address):
                if atype != 'all':
                    addr['type'] = atype
                elif 'atype' in addr and 'type' not in addr:
                    addr['type'] = addr['atype']
                addrObj = Address()
                addrObj.values = addr
                addr = addrObj

            self.append(addr)

    def del_multi(self):
        res = list(filter(_addr_filter(atype), self))
        for addr in res:
            self.iterables.remove(addr)
            self.xml.remove(addr.xml)

    return get_multi, set_multi, del_multi


for atype in ('all', 'bcc', 'cc', 'noreply', 'replyroom', 'replyto', 'to'):
    get_multi, set_multi, del_multi = _build_methods(atype)

    Addresses.interfaces.add(atype)
    setattr(Addresses, "get_%s" % atype, get_multi)
    setattr(Addresses, "set_%s" % atype, set_multi)
    setattr(Addresses, "del_%s" % atype, del_multi)

    if atype == 'all':
        Addresses.interfaces.add('addresses')
        setattr(Addresses, "get_addresses", get_multi)
        setattr(Addresses, "set_addresses", set_multi)
        setattr(Addresses, "del_addresses", del_multi)



register_stanza_plugin(Addresses, Address, iterable=True)



# Slixmpp: The Slick XMPP Library
# Copyright (C) 2011 Nathanael C. Fritz, Lance J.T. Stout
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
import logging

from slixmpp.xmlstream import ElementBase, ET


class FormField(ElementBase):
    namespace = 'jabber:x:data'
    name = 'field'
    plugin_attrib = 'field'
    plugin_multi_attrib = 'fields'
    interfaces = {'answer', 'desc', 'required', 'value',
                  'label', 'type', 'var'}
    sub_interfaces = {'desc'}
    plugin_tag_map = {}
    plugin_attrib_map = {}

    field_types = {'boolean', 'fixed', 'hidden', 'jid-multi',
                   'jid-single', 'list-multi', 'list-single',
                   'text-multi', 'text-private', 'text-single'}

    true_values = {True, '1', 'true'}
    option_types = {'list-multi', 'list-single'}
    multi_line_types = {'hidden', 'text-multi'}
    multi_value_types = {'hidden', 'jid-multi',
                         'list-multi', 'text-multi'}

    def setup(self, xml=None):
        if ElementBase.setup(self, xml):
            self._type = None
        else:
            self._type = self['type']

    def set_type(self, value):
        self._set_attr('type', value)
        if value:
            self._type = value

    def add_option(self, label='', value=''):
        if self._type is None or self._type in self.option_types:
            opt = FieldOption()
            opt['label'] = label
            opt['value'] = value
            self.append(opt)
        else:
            raise ValueError("Cannot add options to " + \
                             "a %s field." % self['type'])

    def del_options(self):
        optsXML = self.xml.findall('{%s}option' % self.namespace)
        for optXML in optsXML:
            self.xml.remove(optXML)

    def del_required(self):
        reqXML = self.xml.find('{%s}required' % self.namespace)
        if reqXML is not None:
            self.xml.remove(reqXML)

    def del_value(self):
        valsXML = self.xml.findall('{%s}value' % self.namespace)
        for valXML in valsXML:
            self.xml.remove(valXML)

    def get_answer(self):
        return self['value']

    def get_options(self):
        options = []
        optsXML = self.xml.findall('{%s}option' % self.namespace)
        for optXML in optsXML:
            opt = FieldOption(xml=optXML)
            options.append({'label': opt['label'], 'value': opt['value']})
        return options

    def get_required(self):
        reqXML = self.xml.find('{%s}required' % self.namespace)
        return reqXML is not None

    def get_value(self, convert=True, convert_list=False):
        """
        Gets the value for this field

        :param convert: Convert truthy values to boolean
        :param convert_list: Convert text-multi fields to a string with
            \n as separator for values
        """
        valsXML = self.xml.findall('{%s}value' % self.namespace)
        if len(valsXML) == 0:
            return None
        elif self._type == 'boolean':
            if convert:
                return valsXML[0].text in self.true_values
            return valsXML[0].text
        elif self._type in self.multi_value_types or len(valsXML) > 1:
            values = []
            for valXML in valsXML:
                if valXML.text is None:
                    valXML.text = ''
                values.append(valXML.text)
            if self._type == 'text-multi' and convert_list:
                values = "\n".join(values)
            return values
        else:
            if valsXML[0].text is None:
                return ''
            return valsXML[0].text

    def set_answer(self, answer):
        self['value'] = answer

    def set_false(self):
        self['value'] = False

    def set_options(self, options):
        for value in options:
            if isinstance(value, dict):
                self.add_option(**value)
            else:
                self.add_option(value=value)

    def set_required(self, required):
        exists = self['required']
        if not exists and required:
            self.xml.append(ET.Element('{%s}required' % self.namespace))
        elif exists and not required:
            del self['required']

    def set_true(self):
        self['value'] = True

    def set_value(self, value):
        del self['value']
        valXMLName = '{%s}value' % self.namespace

        if not self._type:
            if isinstance(value, bool):
                log.debug("Passed a 'boolean' as value of an untyped field, assuming it is a 'boolean'")
                self._type = "boolean"
            elif isinstance(value, str):
                log.debug("Passed a 'str' as value of an untyped field, assuming it is a 'text-single'")
                self._type = "text-single"
            elif isinstance(value, (list, tuple)):
                log.debug("Passed a %s as value of an untyped field, assuming it is a 'text-multi'")
                self._type = "text-multi"

        if self._type == 'boolean':
            if value in self.true_values:
                valXML = ET.Element(valXMLName)
                valXML.text = '1'
                self.xml.append(valXML)
            else:
                valXML = ET.Element(valXMLName)
                valXML.text = '0'
                self.xml.append(valXML)
        elif self._type in self.multi_value_types or self._type in ('', None):
            if isinstance(value, bool):
                value = [value]
            if not isinstance(value, list):
                value = value.replace('\r', '')
                value = value.split('\n')
            for val in value:
                if self._type in ('', None) and val in self.true_values:
                    val = '1'
                valXML = ET.Element(valXMLName)
                valXML.text = val
                self.xml.append(valXML)
        else:
            if isinstance(value, list):
                raise ValueError("Cannot add multiple values " + \
                                 "to a %s field." % self._type)
            valXML = ET.Element(valXMLName)
            valXML.text = value
            self.xml.append(valXML)


class FieldOption(ElementBase):
    namespace = 'jabber:x:data'
    name = 'option'
    plugin_attrib = 'option'
    interfaces = {'label', 'value'}
    sub_interfaces = {'value'}
    plugin_multi_attrib = 'options'


FormField.addOption = FormField.add_option
FormField.delOptions = FormField.del_options
FormField.delRequired = FormField.del_required
FormField.delValue = FormField.del_value
FormField.getAnswer = FormField.get_answer
FormField.getOptions = FormField.get_options
FormField.getRequired = FormField.get_required
FormField.getValue = FormField.get_value
FormField.setAnswer = FormField.set_answer
FormField.setFalse = FormField.set_false
FormField.setOptions = FormField.set_options
FormField.setRequired = FormField.set_required
FormField.setTrue = FormField.set_true
FormField.setValue = FormField.set_value


log = logging.getLogger(__name__)




# Slixmpp: The Slick XMPP Library
# Copyright (C) 2012 Nathanael C. Fritz, Lance J.T. Stout
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
from slixmpp.stanza import Message, Presence, Iq
from slixmpp.xmlstream import ElementBase


class Forwarded(ElementBase):
    name = 'forwarded'
    namespace = 'urn:xmpp:forward:0'
    plugin_attrib = 'forwarded'
    interfaces = {'stanza'}

    def get_stanza(self):
        for stanza in self:
            if isinstance(stanza, (Message, Presence, Iq)):
                return stanza
        return ''

    def set_stanza(self, value):
        self.del_stanza()
        self.append(value)

    def del_stanza(self):
        found_stanzas = []
        for stanza in self:
            if isinstance(stanza, (Message, Presence, Iq)):
                found_stanzas.append(stanza)
        for stanza in found_stanzas:
            self.iterables.remove(stanza)
            self.xml.remove(stanza.xml)




# Slixmpp: The Slick XMPP Library
# Copyright (C) 2010  Nathanael C. Fritz
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
from slixmpp.stanza.error import Error
from slixmpp.xmlstream import StanzaBase, ET


class StreamError(Error, StanzaBase):

    """
    XMPP stanzas of type 'error' should include an <error> stanza that
    describes the nature of the error and how it should be handled.

    Use the 'XEP-0086: Error Condition Mappings' plugin to include error
    codes used in older XMPP versions.

    The stream:error stanza is used to provide more information for
    error that occur with the underlying XML stream itself, and not
    a particular stanza.

    Note: The StreamError stanza is mostly the same as the normal
          Error stanza, but with different namespaces and
          condition names.

    Example error stanza:
        <stream:error>
          <not-well-formed xmlns="urn:ietf:params:xml:ns:xmpp-streams" />
          <text xmlns="urn:ietf:params:xml:ns:xmpp-streams">
            XML was not well-formed.
          </text>
        </stream:error>

    Stanza Interface:
        condition -- The name of the condition element.
        text      -- Human readable description of the error.

    Attributes:
        conditions   -- The set of allowable error condition elements.
        condition_ns -- The namespace for the condition element.

    Methods:
        setup         -- Overrides ElementBase.setup.
        get_condition -- Retrieve the name of the condition element.
        set_condition -- Add a condition element.
        del_condition -- Remove the condition element.
        get_text      -- Retrieve the contents of the <text> element.
        set_text      -- Set the contents of the <text> element.
        del_text      -- Remove the <text> element.
    """

    namespace = 'http://etherx.jabber.org/streams'
    interfaces = {'condition', 'text', 'see_other_host'}
    conditions = {
        'bad-format', 'bad-namespace-prefix', 'conflict',
        'connection-timeout', 'host-gone', 'host-unknown',
        'improper-addressing', 'internal-server-error', 'invalid-from',
        'invalid-namespace', 'invalid-xml', 'not-authorized',
        'not-well-formed', 'policy-violation', 'remote-connection-failed',
        'reset', 'resource-constraint', 'restricted-xml', 'see-other-host',
        'system-shutdown', 'undefined-condition', 'unsupported-encoding',
        'unsupported-feature', 'unsupported-stanza-type',
        'unsupported-version'}
    condition_ns: str = 'urn:ietf:params:xml:ns:xmpp-streams'

    def get_see_other_host(self) -> str | dict[str, str]:
        ns = self.condition_ns
        return self._get_sub_text('{%s}see-other-host' % ns, '')

    def set_see_other_host(self, value: str) -> ET.Element | None:
        if value:
            del self['condition']
            ns = self.condition_ns
            return self._set_sub_text('{%s}see-other-host' % ns, value)
        elif self['condition'] == 'see-other-host':
            del self['condition']
        return None

    def del_see_other_host(self) -> None:
        self._del_sub('{%s}see-other-host' % self.condition_ns)



# Slixmpp: The Slick XMPP Library
# Copyright © 2026 nicoco <nicoco@nicoco.fr>
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.

from slixmpp.plugins.base import register_plugin

from . import stanza
from .mav import XEP_0463

register_plugin(XEP_0463)

__all__ = ("stanza", "XEP_0463")




# Slixmpp: The Slick XMPP Library
# Copyright (C) 2012 Nathanael C. Fritz, Lance J.T. Stout
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
import logging

from slixmpp import Presence
from slixmpp.exceptions import XMPPError
from slixmpp.plugins import BasePlugin, register_plugin
from slixmpp.xmlstream import register_stanza_plugin

from slixmpp.plugins.xep_0012 import stanza, LastActivity


log = logging.getLogger(__name__)


class XEP_0256(BasePlugin):

    name = 'xep_0256'
    description = 'XEP-0256: Last Activity in Presence'
    dependencies = {'xep_0012'}
    stanza = stanza
    default_config = {
        'auto_last_activity': False
    }

    def plugin_init(self):
        register_stanza_plugin(Presence, LastActivity)

        self.xmpp.add_filter('out', self._initial_presence_activity)
        self.xmpp.add_event_handler('connected', self._reset_presence_activity)

        self._initial_presence = set()

    def plugin_end(self):
        self.xmpp.del_filter('out', self._initial_presence_activity)
        self.xmpp.del_event_handler('connected', self._reset_presence_activity)

    def _reset_presence_activity(self, e):
        self._initial_presence = set()

    def _initial_presence_activity(self, stanza):
        if isinstance(stanza, Presence):
            use_last_activity = False

            if self.auto_last_activity and  stanza['show'] in ('xa', 'away'):
                use_last_activity = True

            if stanza['from'] not in self._initial_presence:
                self._initial_presence.add(stanza['from'])
                use_last_activity = True

            if use_last_activity:
                plugin = self.xmpp['xep_0012']
                try:
                    result = plugin.api['get_last_activity'](stanza['from'],
                                                             None,
                                                             stanza['to'])
                    seconds = result['last_activity']['seconds']
                except XMPPError:
                    seconds = None

                if seconds is not None:
                    stanza['last_activity']['seconds'] = seconds
        return stanza


register_plugin(XEP_0256)
